#include "typedef.h"
#include <stdio.h>
#include <string.h>

#include "util.h"
#include "object.h"
#include "directory.h"
#include "class.h"

#ifdef WIN32
#include <malloc.h>
#define alloca _alloca
#else
#include <alloca.h>
#endif

Cdirectory	*Cdirectory::g_headP     = 0;
Cdirectory	*Cdirectory::g_tailP     = 0;
Cdirectory	*Cdirectory::g_collapseP = 0;

void
Cdirectory0::name(const char *nameP)
{
	if (m_nameP) {
		xfree(m_nameP);
	}

	m_nameP = (char *) xmalloc(strlen(nameP) + 1);
	strcpy(m_nameP, nameP);
}


Cdirectory0 *
Cdirectory0::locateForClass(const Cclass *headP, u2T *nameP)
{
	const Cclass	*classP;
	u1T				*bufferP, *P;
	u2T				*P2;
	u2T				c;

	for (classP = headP; classP; classP = classP->m_hashP) {
		if (classP->m_parentP) {
			if (!unicodecmp(classP->m_nameP, nameP)) {
				return ((Cdirectory0 *) classP->m_parentP);
	}	}	}
	
	for (P2 = nameP; c = *P2; ++P2) {
		if (c > 255) {
			return(0);
	}	}
	bufferP = P = (u1T *) alloca(unicodelth(nameP) + 1);
	for (P2 = nameP; *P++ = (u1T) *P2++; );
	return Cdirectory::locateForClass((char *) bufferP);
}

Cdirectory::Cdirectory(Centity *parentP, const char *filenameP)
{
	m_nextP     = 0;
	if (!g_headP) {
		g_headP = this;
	} else {
		g_tailP->m_nextP = this;
	}
	g_tailP = this;

	m_nameP    = 0;
	m_parentP  = parentP;
	m_collapse = 0;
	name(filenameP);
	add_created();
}

Cdirectory *
Cdirectory::locateForClass(char *nameP)
{
	Cdirectory		*directoryP;
	Cdirectory		*directory1P;
	char			*P,*P1, *P2, *lastP;
	char			c;

	directoryP = 0;
	P1         = 0;
	for (P = nameP; c = *P; ++P) {
		if (c == '/') {
			P1 = P;
	}	}

	if (lastP = P1) {	// Has last slash
		*P1 = 0;
		for (P2 = P1; P2 > nameP; --P2) {
			if (*P2 == '/') {
				++P2;
				break;
		}	}
		for (directoryP = Cdirectory::g_headP; directoryP; directoryP = directoryP->m_nextP) {
			if (strcmp(P2, directoryP->m_nameP)) {
				continue;
			}
			// Last part of name matches directory
			for (directory1P = directoryP; P2 > nameP && (directory1P = ((Cdirectory *) directory1P->m_parentP)); ) {
				*P1 = '/';
				P1  = --P2;
				*P1 = 0;
				for (P2 = P1; P2 > nameP; --P2) {
					if (*P2 == '/') {
						++P2;
						break;
				}	}
				if (strcmp(P2, directory1P->m_nameP)) {
					break;
			}	}
			if (P2 == nameP) {
				*lastP = '/';
				return(directoryP);
		}	}
		*P1 = '/';
		*lastP = 0;
		directory1P = locateForClass(nameP);
		for (P2 = lastP; P2 > nameP; --P2) {
			if (*P2 == '/') {
				++P2;
				break;
		}	}
		directoryP = new Cdirectory(directory1P, P2);
		*lastP = '/';
		if (!directoryP) {
			outofmemory();
		}
	}
	return(directoryP);
}

/* Find the directory at the bottom of  the directory path pathP 
 * whose root directory occurs under rootP
 */

Centity *
Cdirectory::locate(char *pathP, Centity *rootP)
{
	Centity		*leafP;
	char		*dirnameP;
	char		*endP;
	Cdirectory	*fromP;

	leafP    = rootP;
	dirnameP = pathP;
descend:
	endP = strchr(dirnameP, '/');
	if (!endP) {
		goto done;
	}
	*endP = 0;
	for (fromP = g_headP; fromP; fromP = fromP->m_nextP) {
		if (fromP->m_parentP == leafP  && !strcmp(fromP->m_nameP, dirnameP)) {
			leafP    = fromP;
			*endP    = '/';
			dirnameP = endP+1;
			goto descend;
	}	}

	// Directory not found

	for (;;) {
		leafP = new Cdirectory(leafP, dirnameP);
		*endP    = '/';
		dirnameP = endP+1;
		endP     = strchr(dirnameP, '/');
		if (!endP) {
			break;
		}
		*endP = 0;
	}
done:
	return(leafP);
}

size_t
Cdirectory::full_path(char *bufferP) const
{
	size_t	lth;

	if (m_parentP) {
		lth = m_parentP->full_path(bufferP);
	} else {
		lth = 0;
	}
	strcpy(bufferP+lth, m_nameP);
	lth += strlen(m_nameP);
	bufferP[lth++] = '/';
	return(lth);
}

/* virtual */ objectE
Cdirectory::type(void) const
{
	return(directoryE);
}

/* virtual */ void
Cdirectory::emit_id(void) const
{
	fprintf(stdout, "D%p", this);
}

/* virtual */ void
Cdirectory::emit_instance(void) const
{
	if (m_collapse != -2) {
		fprintf(stdout, "$INSTANCE D%p D\n", this);
}	}

/* virtual */ void
Cdirectory::emit_edges(void)
{
	if (m_collapse != -2) {
		Centity::emit_edges();
}	}

void
Cdirectory::emit_attributes(void) const
{
	const char *nameP;
	const char *prefixP;

	if (m_collapse != -2) {
		nameP = m_nameP;
		if (!*m_nameP) {
			nameP = "/";
		}
		if (*nameP == '/' || nameP[1] == ':') {
			prefixP = "";
		} else {
			prefixP = "./";
		}
		emit_id();
		fprintf(stdout, " {label=\"%s\" file=\"%s%s\"}\n", nameP, prefixP, nameP);
	}
}
