#include "typedef.h"

#include <stdio.h>
#include <string.h>

#include "util.h"
#include "object.h"
#include "directory.h"
#include "class.h"
#include "source.h"

Csource	*Csource::g_hash[HASHSIZE] = {0};

/* static */ u4T
Csource::hash(const char *filenameP)
{
	int		ret  = 0;
	u1T		c;
	const	char *P;

	for (P = filenameP; c = *P ; ++P) {
		ret  += c;
		ret <<= 1;
		if (ret & HASHSIZE) {
			ret |= 1;
	}	}
	return(ret & (HASHSIZE-1));
}

/* Source is preserved in the order visited */

Csource::Csource(char *pathP, const char *filenameP)
{
	m_hashP    = 0;
	
	if (pathP) {
		m_parentP = Cdirectory::locate(pathP, 0);
	}
	m_nameP   = 0;
	name(filenameP);
	add_created();
}

Csource *
Csource::locate(char *pathP)
{
	char			*filenameP, *dirnameP, *P1;
	unsigned int	index;
	Csource			**sourcePP, *sourceP;
	int				c, diff;
	const char		*dirname1P;
	Centity			*parentP;

	filenameP = strrchr(pathP, '/');
	if (filenameP) {
		++filenameP;
	} else {
		filenameP = pathP;
	}

	index = hash(filenameP);
	for (sourcePP = g_hash + index; sourceP = *sourcePP; sourcePP = &(sourceP->m_hashP)) {
		if (strcmp(sourceP->m_nameP, filenameP)) {
			// Different source file name
			continue;
		}
		dirnameP = filenameP;
		for (parentP = sourceP->m_parentP;; parentP = parentP->m_parentP) {
			if (!parentP || parentP->type() != directoryE) {
				if (dirnameP == pathP) {
					return(sourceP);
				}
				break;
			}

			if (dirnameP == pathP) {
				break;
			}
			dirname1P = ((Cdirectory *) parentP)->m_nameP;

			P1 = --dirnameP;
			if (P1 != pathP) {
				/* Otherwise P1="/" ie root represented by "" */

				for (dirnameP = P1 - 1; ; --dirnameP) {
					if (*dirnameP == '/') {
						++dirnameP;
						break;
					}
					if (dirnameP == pathP) {
						break;
					}
			}	}
			c    = *P1;
			*P1  = 0;
			diff = strcmp(dirnameP, dirname1P);
			*P1  = c;
			if (diff) {
				// This source is under a different path from the one desired
				break;
	}	}	}

	/* Didn't find if get here */

	sourceP = new Csource(pathP, filenameP);
	if (!sourceP) {
		outofmemory();
	}
	*sourcePP = sourceP;
	return(sourceP);
}

size_t
Csource::full_path(char *bufferP) const
{
	size_t	lth;

	if (m_parentP) {
		lth = m_parentP->full_path(bufferP);
	} else {
		lth = 0;
	}
	strcpy(bufferP+lth, m_nameP);
	lth += (int) strlen(m_nameP);
	return(lth);
}

/* virtual */ objectE
Csource::type(void) const
{
	return(sourceE);
}

/* virtual */ void
Csource::emit_id(void) const
{
	fprintf(stdout, "S%p", this);
}

/* virtual */ void
Csource::emit_instance(void) const
{
	fprintf(stdout, "$INSTANCE S%p S\n", this);
}

void
Csource::emit_attributes(void) const
{
	emit_id();
	fprintf(stdout, " {label=\"%s\" file=\"./%s\"}\n", m_nameP, m_nameP);
}
